import os
import logging

logger = logging.getLogger(__name__)


def prepare_path(file_target):
    # as the path might not exist, we can not test if it is a dir beforehand
    # therefore, if it ends with a / it is considered a dir, otherwise, it is a regular file
    # and the following code works for both cases
    os.makedirs(os.path.abspath(file_target.rpartition('/')[0]), exist_ok=True)




import os
import codecs
import logging
from tempfile import NamedTemporaryFile

import editor

from .storages import get_storage, strip_prefix
from .utils import prepare_path
from .config import Settings
from . import exceptions, files

logger = logging.getLogger(__name__)
__escape_decoder = codecs.getdecoder('unicode_escape')


def parse_dotenv(data):
    for line in data.splitlines():
        line = line.strip()
        if line and not line.startswith('#') and '=' in line:
            k, _, v = line.partition('=')

            # Remove any leading and trailing spaces in key, value
            k, v = k.strip(), v.strip().encode('unicode-escape').decode('ascii')

            if v and v[0] == v[-1] in ['"', "'"]:
                v = __escape_decoder(v[1:-1])[0]

            yield k, v


def unpack_list(files_list):
    files_pairs = files_list.split(';')
    files_map = []
    for file_map in files_pairs:
        file_source, _, file_target = file_map.rpartition(':')
        if file_source and file_target:
            files_map.append((file_source, file_target))
    return files_map


def phusion_dump(environment, path):
    prepare_path(path if path.endswith('/') else path + '/')
    for k, v in environment.items():
        with open(os.path.join(path, k), 'w') as f:
            f.write(v + '\n')


def setup_environment(storage='s3',
                      dump=False,
                      dump_path='/etc/container_environment',
                      **kwargs):
    try:
        conf = S3Conf(storage=storage)
        env_vars = conf.get_envfile(**kwargs).as_dict()
        for var_name, var_value in env_vars.items():
            print('{}={}'.format(var_name, var_value))
        if dump:
            phusion_dump(env_vars, dump_path)
    except ValueError as e:
        logger.error(e)
    except Exception as e:
        logger.error(e)
        raise e


class S3Conf:
    def __init__(self, storage='s3', settings=None):
        self.settings = settings or Settings()
        self.storage = get_storage(storage, settings=self.settings)

    @property
    def environment_file_path(self):
        # resolving environment file path
        file_name = self.settings.get('S3CONF')
        if not file_name:
            logger.error('Environemnt file name is not defined or is empty.')
            raise exceptions.EnvfilePathNotDefinedError()
        return file_name

    def downsync(self, files, root_dir=None):
        if isinstance(files, str):
            files = unpack_list(files)
        for remote_file, local_file in files:
            self.download(remote_file, local_file, root_dir=root_dir)

    def upsync(self, files, root_dir=None):
        if isinstance(files, str):
            files = unpack_list(files)
        for remote_file, local_file in files:
            self.upload(local_file, remote_file, root_dir=root_dir)

    def download(self, path, path_target, root_dir=None):
        if root_dir:
            path_target = os.path.join(root_dir, path_target.lstrip('/'))
        logger.info('Downloading %s to %s', path, path_target)
        for file_path in self.storage.list(path):
            if path.endswith('/') or not path:
                target_name = os.path.join(path_target, file_path)
            else:
                target_name = path_target
            prepare_path(target_name)
            with open(target_name, 'wb') as f:
                # join might add a trailing slash, but we know it is a file, so we remove it
                # stream=f reads the data into f and returns f as our open file
                self.storage.open(os.path.join(path, file_path).rstrip('/'), stream=f)

    def upload(self, path, path_target, root_dir=None):
        if root_dir:
            path = os.path.join(root_dir, path.lstrip('/'))
        logger.info('Uploading %s to %s', path, path_target)
        if os.path.isdir(path):
            for root, dirs, files in os.walk(path):
                for file in files:
                    file_source = os.path.join(root, file)
                    file_target = os.path.join(path_target, strip_prefix(os.path.join(root, file), path).lstrip('/'))
                    self.storage.write(open(file_source, 'rb'), file_target)
        else:
            self.storage.write(open(path, 'rb'), path_target)

    def get_envfile(self):
        logger.info('Loading configs from {}'.format(self.environment_file_path))
        return files.EnvFile(self.environment_file_path, storage=self.storage)

    def edit(self):
        with NamedTemporaryFile(mode='rb+', buffering=0) as f:
            original_data = self.storage.open(self.environment_file_path).read()
            f.write(original_data)
            edited_data = editor.edit(filename=f.name)
            if edited_data != original_data:
                self.upload(f.name, self.environment_file_path)
            else:
                logger.warning('File not changed. Nothing to upload.')



import logging
import io
import codecs
from tempfile import NamedTemporaryFile

import editor

from .storages import LocalStorage

logger = logging.getLogger(__name__)
__escape_decoder = codecs.getdecoder('unicode_escape')


class File:
    def __init__(self, name, storage=None):
        self.name = name
        self.storage = storage or LocalStorage()
        self._stream = None

    def read(self):
        return self.storage.open(self.name).read()

    def write(self, data):
        if isinstance(data, str):
            data = data.encode('utf-8')
        self.storage.write(io.BytesIO(data), self.name)

    def edit(self):
        with NamedTemporaryFile(mode='rb+', buffering=0) as f:
            original_data = self.read()
            f.write(original_data)
            edited_data = editor.edit(filename=f.name)
            if edited_data != original_data:
                f.seek(0)
                self.storage.write(f, self.name)
            else:
                logger.warning('File not changed. Nothing to write.')


class EnvFile(File):
    def as_dict(self):
        env_dict = {}
        try:
            for line in str(self.read(), 'utf-8').splitlines():
                line = line.strip()
                if line and not line.startswith('#') and '=' in line:
                    k, _, v = line.partition('=')

                    # Remove any leading and trailing spaces in key, value
                    k, v = k.strip(), v.strip().encode('unicode-escape').decode('ascii')

                    if v and v[0] == v[-1] in ['"', "'"]:
                        v = __escape_decoder(v[1:-1])[0]

                    env_dict[k] = v
        except FileNotFoundError:
            pass
        return env_dict

    def from_dict(self, env_vars):
        file_data = ''
        for var_name, var_value in env_vars.items():
            file_data += '{}={}\n'.format(var_name, var_value)
        self.write(file_data)

    def update(self, another_file):
        env_vars = self.as_dict()
        env_vars.update(another_file.as_dict())
        self.from_dict(env_vars)



import os
import logging
import boto3
from io import BytesIO

from .utils import prepare_path
from .config import Settings


logger = logging.getLogger(__name__)


def strip_prefix(text, prefix):
    return text[len(prefix):] if text.startswith(prefix) else text


def strip_s3_path(path):
    bucket, _, path = strip_prefix(path, 's3://').partition('/')
    return bucket, path


class BaseStorage:
    def __init__(self, settings=None):
        self._settings = settings or Settings()

    def open(self, file_name, stream=None):
        raise NotImplementedError()

    def write(self, f, file_name):
        raise NotImplementedError()

    def list(self, path):
        raise NotImplementedError()


class S3Storage(BaseStorage):
    def __init__(self, settings=None):
        super(__class__, self).__init__(settings=settings)
        self._resource = None

    def get_resource(self):
        logger.debug('Getting S3 resource')
        # See how boto resolve credentials in
        # http://boto3.readthedocs.io/en/latest/guide/configuration.html#guide-configuration
        if not self._resource:
            logger.debug('Resource does not exist, creating a new one...')
            self._resource = boto3.resource(
                's3',
                aws_access_key_id=self._settings.get('S3CONF_ACCESS_KEY_ID') or self._settings.get('AWS_ACCESS_KEY_ID'),
                aws_secret_access_key=self._settings.get('S3CONF_SECRET_ACCESS_KEY') or self._settings.get('AWS_SECRET_ACCESS_KEY'),
                aws_session_token=self._settings.get('S3CONF_SESSION_TOKEN') or self._settings.get('AWS_SESSION_TOKEN'),
                region_name=self._settings.get('S3CONF_S3_REGION_NAME') or self._settings.get('AWS_S3_REGION_NAME'),
                use_ssl=self._settings.get('S3CONF_S3_USE_SSL') or self._settings.get('AWS_S3_USE_SSL', True),
                endpoint_url=self._settings.get('S3CONF_S3_ENDPOINT_URL') or self._settings.get('AWS_S3_ENDPOINT_URL'),
            )
        return self._resource

    def _read_file_into_stream(self, bucket, file_name, stream=None):
        s3 = self.get_resource()
        stream = stream or BytesIO()
        s3.Object(bucket, file_name).download_fileobj(stream)
        stream.seek(0)
        return stream

    def _write_file(self, f, bucket, path_target):
        s3 = self.get_resource()
        s3.Object(bucket, path_target).upload_fileobj(f)

    # this is not good, it should return a file like and not mix
    # a "read-into" functionality with the "open" behavior
    # but it works for now, must fix this at some point
    def open(self, file_name, stream=None):
        logger.debug('Reading from %s', file_name)
        bucket, file_name = strip_s3_path(file_name)
        return self._read_file_into_stream(bucket, file_name, stream=stream)

    def write(self, f, file_name):
        logger.debug('Writing to %s', file_name)
        bucket, path_target = strip_s3_path(file_name)
        self._write_file(f, bucket, path_target)

    def list(self, path):
        logger.debug('Listing %s', path)
        bucket_name, path = strip_s3_path(path)
        bucket = self.get_resource().Bucket(bucket_name)
        for obj in bucket.objects.filter(Prefix=path):
            if not obj.key.endswith('/'):
                yield strip_prefix(obj.key, path)


class LocalStorage(BaseStorage):
    def _validate_path(self, path):
        if path.startswith('s3://'):
            raise ValueError('LocalStorage can not process S3 paths.')

    def open(self, file_name, stream=None):
        self._validate_path(file_name)
        stream = stream or BytesIO()
        with open(file_name, 'rb') as f:
            stream.write(f.read())
        stream.seek(0)
        return stream

    def write(self, f, file_name):
        self._validate_path(file_name)
        prepare_path(file_name)
        open(file_name, 'wb').write(f.read())

    def list(self, path):
        self._validate_path(path)
        if os.path.isdir(path):
            for root, dirs, files in os.walk(path):
                for file in files:
                    yield strip_prefix(os.path.join(root, file), path)
        else:
            # the relative path of a file to itself is empty
            yield ''


STORAGES = {
    's3': S3Storage,
    'local': LocalStorage,
}


def get_storage(storage, settings=None):
    if not isinstance(storage, BaseStorage):
        return STORAGES[storage](settings=settings)
    return storage



import logging
import subprocess
import shlex
import os
from shutil import rmtree

import click
from click.exceptions import UsageError
import click_log

from . import s3conf, config, files, exceptions


logger = logging.getLogger(__name__)


def get_settings(section=None):
    if section:
        section = config.Settings(section=section)
    else:
        section = config.Settings()
    return section


@click.group(invoke_without_command=True)
@click.version_option()
@click.option('--edit', '-e', is_flag=True)
@click.option('--global', 'global_settings', is_flag=True)
@click.pass_context
# this sets the log level for this app only
@click_log.simple_verbosity_option('s3conf')
def main(ctx, edit, global_settings):
    """
    Simple command line tool to help manage environment variables stored in a S3-like system. Facilitates editing text
    files remotely stored, as well as downloading and uploading files.
    """
    # configs this module logger to behave properly
    # logger messages will go to stderr (check __init__.py/patch.py)
    # client output should be generated with click.echo() to go to stdout
    click_log.basic_config('s3conf')
    logger.debug('Running main entrypoint')
    if edit:
        if ctx.invoked_subcommand is None:
            if global_settings:
                logger.debug('Using config file %s', config.GLOBAL_CONFIG_FILE)
                config.ConfigFileResolver(config.GLOBAL_CONFIG_FILE).edit()
            else:
                logger.debug('Using config file %s', config.LOCAL_CONFIG_FILE)
                config.ConfigFileResolver(config.LOCAL_CONFIG_FILE).edit()
            return
        else:
            raise UsageError('Edit should not be called with a subcommand.')
    # manually call help in case no relevant settings were defined
    if ctx.invoked_subcommand is None:
        click.echo(main.get_help(ctx))


@main.command('env')
@click.argument('section',
                required=False)
@click.option('--storage',
              type=click.Choice(['s3', 'local']),
              default='s3',
              show_default=True,
              help='Storage driver to use. Local driver is mainly for testing purpouses.')
@click.option('--map-files',
              '-m',
              is_flag=True,
              help='If defined, tries to map files from the storage to the local drive as defined by '
                   'the variable S3CONF_MAP read from the S3CONF file.')
@click.option('--phusion',
              is_flag=True,
              help='If set, dumps variables to --dump-path in for format used by the phusion docker image. '
                   'More information in https://github.com/phusion/baseimage-docker.')
@click.option('--phusion-path',
              default='/etc/container_environment',
              show_default=True,
              help='Path where to dump variables as in the phusion docker image format.')
@click.option('--quiet',
              '-q',
              is_flag=True,
              help='Do not print any environment variables output.')
@click.option('--edit',
              '-e',
              is_flag=True)
def env(section, storage, map_files, phusion, phusion_path, quiet, edit):
    """
    Reads the file defined by the S3CONF variable and output its contents to stdout. Logs are printed to stderr.
    See options for added functionality: editing file, mapping files, dumping in the phusion-baseimage format, etc.
    """
    try:
        logger.debug('Running env command')
        settings = get_settings(section)
        conf = s3conf.S3Conf(storage=storage, settings=settings)

        if edit:
            conf.edit()
        else:
            env_vars = conf.get_envfile().as_dict()
            if env_vars.get('S3CONF_MAP') and map_files:
                conf.downsync(env_vars.get('S3CONF_MAP'))
            if not quiet:
                for var_name, var_value in sorted(env_vars.items(), key=lambda x: x[0]):
                    click.echo('{}={}'.format(var_name, var_value))
            if phusion:
                s3conf.phusion_dump(env_vars, phusion_path)
    except exceptions.EnvfilePathNotDefinedError:
        error_msg = 'Set the environemnt variable S3CONF or provide a section from an existing config file.'
        try:
            sections_detected = ''
            for section in config.ConfigFileResolver(config.LOCAL_CONFIG_FILE).sections():
                sections_detected += '    {}\n'.format(section)
        except FileNotFoundError:
            pass
        if sections_detected:
            sections_detected = '\n\nThe following sections were detected:\n\n' + sections_detected
        raise UsageError(error_msg + sections_detected)


@main.command('exec')
@click.option('--storage',
              type=click.Choice(['s3', 'local']),
              default='s3',
              show_default=True,
              help='Storage driver to use. Local driver is mainly for testing purpouses.')
@click.option('--map-files',
              '-m',
              is_flag=True,
              help='If defined, tries to map files from the storage to the local drive as defined by '
                   'the variable S3CONF_MAP read from the S3CONF file.')
@click.argument('section',
                required=False)
# to use option-like arguments use "--"
# http://click.pocoo.org/6/arguments/#option-like-arguments
@click.argument('command',
                required=False,
                nargs=-1,
                type=click.UNPROCESSED)
@click.pass_context
def exec_command(ctx, section, command, storage, map_files):
    """
    Sets the process environemnt and executes the [COMMAND] in the same context. Does not modify the current shell
    environment.

    If the [COMMAND] has option-like arguments, use the standard POSIX pattern "--" to separate options
    from arguments. Considering our configuration in the "dev" section, we could write:

    s3conf -v info exec dev -- ping -v google.com
    """
    try:
        logger.debug('Running exec command')
        existing_sections = config.ConfigFileResolver(config.LOCAL_CONFIG_FILE).sections()
        command = ' '.join(command)
        if section not in existing_sections:
            command = '{} {}'.format(section, command) if command else section
            section = None

        if not command:
            logger.warning('No command detected.')
            click.echo(exec_command.get_help(ctx))
            return

        settings = get_settings(section)
        conf = s3conf.S3Conf(storage=storage, settings=settings)

        env_vars = conf.get_envfile().as_dict()
        if env_vars.get('S3CONF_MAP') and map_files:
            conf.downsync(env_vars.get('S3CONF_MAP'))

        current_env = os.environ.copy()
        current_env.update(env_vars)
        logger.debug('Executing command "%s"', command)
        subprocess.run(shlex.split(command), env=current_env, check=True)
    except exceptions.EnvfilePathNotDefinedError:
        error_msg = 'Set the environemnt variable S3CONF or provide a section from an existing config file.'
        try:
            sections_detected = ''
            for section in existing_sections:
                sections_detected += '    {}\n'.format(section)
        except FileNotFoundError:
            pass
        if sections_detected:
            sections_detected = '\n\nThe following sections were detected:\n\n' + sections_detected
        raise UsageError(error_msg + sections_detected)


@main.command('download')
@click.argument('remote_path')
@click.argument('local_path')
@click.option('--storage',
              type=click.Choice(['s3', 'local']),
              default='s3',
              show_default=True,
              help='Storage driver to use. Local driver is mainly for testing purpouses.')
def download(remote_path, local_path, storage):
    """
    Download a file or folder from the S3-like service.

    If REMOTE_PATH has a trailing slash it is considered to be a folder, e.g.: "s3://my-bucket/my-folder/". In this
    case, LOCAL_PATH must be a folder as well. The files and subfolder structure in REMOTE_PATH are copied to
    LOCAL_PATH.

    If REMOTE_PATH does not have a trailing slash, it is considered to be a file, and LOCAL_PATH should be a file as
    well.
    """
    conf = s3conf.S3Conf(storage=storage)
    conf.download(remote_path, local_path)


@main.command('upload')
@click.argument('local_path')
@click.argument('remote_path')
@click.option('--storage',
              type=click.Choice(['s3', 'local']),
              default='s3',
              show_default=True,
              help='Storage driver to use. Local driver is mainly for testing purpouses.')
def upload(remote_path, local_path, storage):
    """
    Upload a file or folder to the S3-like service.

    If LOCAL_PATH is a folder, the files and subfolder structure in LOCAL_PATH are copied to REMOTE_PATH.

    If LOCAL_PATH is a file, the REMOTE_PATH file is created with the same contents.
    """
    conf = s3conf.S3Conf(storage=storage)
    conf.upload(local_path, remote_path)


@main.command('downsync')
@click.option('--storage',
              type=click.Choice(['s3', 'local']),
              default='s3',
              show_default=True,
              help='Storage driver to use. Local driver is mainly for testing purpouses.')
@click.option('--map-files',
              '-m',
              is_flag=True,
              help='If defined, tries to map files from the storage to the local drive as defined by '
                   'the variable S3CONF_MAP read from the S3CONF file using the config folder as the '
                   'root directory.')
def downsync(storage, map_files):
    """
    For each section defined in the local config file, creates a folder inside the local config folder
    named after the section. Downloads the environemnt file defined by the S3CONF variable for this section
    to this folder.
    """
    local_resolver = config.ConfigFileResolver(config.LOCAL_CONFIG_FILE)

    for section in local_resolver.sections():
        settings = get_settings(section=section)
        # removing environment resolver, we only want config file onwards on our chain
        settings.resolvers = settings.resolvers[1:]

        # preparing paths
        s3conf_env_file = settings['S3CONF']
        local_root = os.path.join(config.LOCAL_CONFIG_FOLDER, section)
        rmtree(local_root, ignore_errors=True)

        # running operations
        local_path = os.path.join(local_root, os.path.basename(s3conf_env_file).lstrip('/'))
        conf = s3conf.S3Conf(storage=storage, settings=settings)
        remote_env_file = conf.get_envfile()
        local_env_file = files.EnvFile(local_path)
        local_env_file.write(remote_env_file.read())

        if map_files:
            env_vars = local_env_file.as_dict()
            local_mapping_root = os.path.join(local_root, 'root')
            if env_vars.get('S3CONF_MAP'):
                conf.downsync(env_vars.get('S3CONF_MAP'), root_dir=local_mapping_root)


@main.command('upsync')
@click.option('--storage',
              type=click.Choice(['s3', 'local']),
              default='s3',
              show_default=True,
              help='Storage driver to use. Local driver is mainly for testing purpouses.')
@click.option('--map-files',
              '-m',
              is_flag=True,
              help='If defined, tries to map files from the storage to the local drive as defined by '
                   'the variable S3CONF_MAP read from the S3CONF file using the config folder as the '
                   'root directory.')
def upsync(storage, map_files):
    """
    For each section defined in the local config file, look up for a folder inside the local config folder
    named after the section. Uploads the environemnt file named as in the S3CONF variable for this section
    to the remote S3CONF path.
    """
    local_resolver = config.ConfigFileResolver(config.LOCAL_CONFIG_FILE)

    for section in local_resolver.sections():
        settings = get_settings(section=section)
        # removing environment resolver, we only want config file onwards on our chain
        settings.resolvers = settings.resolvers[1:]

        # preparing paths
        s3conf_env_file = settings['S3CONF']
        local_root = os.path.join(config.LOCAL_CONFIG_FOLDER, section)

        # running operations
        local_path = os.path.join(local_root, os.path.basename(s3conf_env_file).lstrip('/'))
        conf = s3conf.S3Conf(storage=storage, settings=settings)
        remote_env_file = conf.get_envfile()
        local_env_file = files.EnvFile(local_path)
        remote_env_file.write(local_env_file.read())

        if map_files:
            env_vars = local_env_file.as_dict()
            local_mapping_root = os.path.join(local_root, 'root')
            if env_vars.get('S3CONF_MAP'):
                conf.upsync(env_vars.get('S3CONF_MAP'), root_dir=local_mapping_root)


if __name__ == '__main__':
    main()



import os
import logging
from tempfile import NamedTemporaryFile
from configparser import ConfigParser

import editor

from .utils import prepare_path


logger = logging.getLogger(__name__)


GLOBAL_CONFIG_FILE = '~/.s3conf'
LOCAL_CONFIG_FOLDER = './.s3conf'
LOCAL_CONFIG_FILE = os.path.join(LOCAL_CONFIG_FOLDER, 'config')


class EnvironmentResolver:
    def get(self, item, default=None):
        return os.environ.get(item, default)


class ConfigFileResolver:
    def __init__(self, config_file, section=None):
        self.config_file = os.path.expanduser(config_file)
        self.section = section or 'DEFAULT'
        self._config = None

    @property
    def config(self):
        if not self._config:
            self._config = ConfigParser()
            self._config.read(self.config_file)
        return self._config

    @config.setter
    def config(self, value):
        self._config = value

    def get(self, item, default=None, section=None):
        return self.config.get(section or self.section, item, fallback=default)

    def edit(self):
        if os.path.isfile(self.config_file):
            editor.edit(filename=self.config_file)
        else:
            with NamedTemporaryFile(mode='rb+', buffering=0) as f:
                data = editor.edit(filename=f.name)

            if data:
                prepare_path(self.config_file)
                with open(self.config_file, 'wb') as f:
                    f.write(data)
            else:
                logger.warning('Nothing to write. Config file not created.')

    def sections(self):
        return self.config.sections()


class Settings:
    def __init__(self, section=None):
        self.resolvers = [
            EnvironmentResolver(),
            ConfigFileResolver(LOCAL_CONFIG_FILE, section),
            ConfigFileResolver(GLOBAL_CONFIG_FILE),
        ]

    def __getitem__(self, item):
        for resolver in self.resolvers:
            value = resolver.get(item)
            if value:
                break
        else:
            logger.debug('Entry %s not found', item)
            raise KeyError()
        logger.debug('Entry %s has value %s', item, value)
        return value

    def get(self, item, default=None):
        try:
            return self[item]
        except KeyError:
            logger.debug('Key %s not found, returning default %s', item, default)
            return default
